@@ -31,7 +31,6 @@ class Sale:
     __name__ = 'sale.sale'
     carrier = fields.Many2One('carrier', 'Carrier',
         domain=[('carrier_product.salable', '=', True)],
-        on_change=['carrier', 'party', 'currency', 'sale_date'],
         states={
             'readonly': Eval('state') != 'draft',
         },
@@ -43,18 +42,6 @@ class Sale:
             'readonly': Eval('state') != 'draft',
             }, depends=['state'])
 
-    @classmethod
-    def __setup__(cls):
-        super(Sale, cls).__setup__()
-
-        for fname in ('carrier', 'party', 'currency', 'sale_date',
-                'shipment_cost_method', 'lines'):
-            if fname not in cls.lines.on_change:
-                cls.lines.on_change.append(fname)
-        for fname in cls.lines.on_change:
-            if fname not in cls.carrier.on_change:
-                cls.carrier.on_change.append(fname)
-
     @staticmethod
     def default_carrier():
         Config = Pool().get('sale.configuration')
@@ -70,9 +57,12 @@ class Sale:
     def _get_carrier_context(self):
         return {}
 
+    @fields.depends(methods=['lines'])
     def on_change_carrier(self):
         return self.on_change_lines()
 
+    @fields.depends('carrier', 'party', 'currency', 'sale_date',
+        'shipment_cost_method', 'lines')
     def on_change_lines(self):
         pool = Pool()
         Product = pool.get('product.product')
@@ -95,16 +85,14 @@ class Sale:
 
         cost_line = {}
         products = [line.product for line in self.lines or []
-                if line.product]
+                if getattr(line, 'product', None)]
         stockable = any(product.type in ('goods', 'assets')
             for product in products)
         if cost and currency_id and stockable:
-            if (self.currency
-                    and currency_id != self.currency.id):
-                date = self.sale_date or today
-                with Transaction().set_context(date=date):
-                    cost = Currency.compute(Currency(currency_id), cost,
-                        self.currency)
+            date = self.sale_date or today
+            with Transaction().set_context(date=date):
+                cost = Currency.compute(Currency(currency_id), cost,
+                    self.currency)
             product = self.carrier.carrier_product
             with Transaction().set_context(party_context):
                 description = Product(product.id).rec_name
@@ -140,12 +128,13 @@ class Sale:
         if not self.lines:
             self.lines = []
         for line in self.lines:
-            if line.shipment_cost:
+            if getattr(line, 'shipment_cost', None):
                 if line.shipment_cost != cost_line.get('shipment_cost'):
                     to_remove = line.id
                     self.lines.remove(line)
                     if cost_line:
-                        cost_line['description'] = line.description
+                        cost_line['description'] = getattr(line, 'description',
+                            '')
                 else:
                     cost_line = {}
                 break
@@ -158,7 +147,7 @@ class Sale:
         if to_remove:
             lines.setdefault('remove', []).append(to_remove)
         if cost_line:
-            lines.setdefault('add', []).append(cost_line)
+            lines.setdefault('add', []).append((-1, cost_line))
         return result
 
     def create_shipment(self, shipment_type):
